"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const typescript_1 = __importDefault(require("typescript"));
/**
 * An ESLint rule to ban unsafe type assertions.
 */
const noUnsafeTypeAssertion = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Bans unsafe type assertions.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringGeneric: "This type assertion is not type-safe.",
            errorStringPreferAs: "Use the 'as' syntax for type assertions.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
        // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/39
        const checker = parserServices.program.getTypeChecker();
        const isUnsafe = (rawDestinationType, rawSourceType) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (rawSourceType.flags & typescript_1.default.TypeFlags.Any ||
                rawSourceType.flags & typescript_1.default.TypeFlags.Unknown) {
                // Asserting any or unknown to anything else is always unsafe.
                return true;
            }
            return (checker.isTypeAssignableTo !== undefined &&
                !checker.isTypeAssignableTo(rawSourceType, rawDestinationType));
        };
        const reportUnsafe = (node
        // eslint-disable-next-line functional/no-return-void
        ) => {
            // The right hand side of the "as".
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const rawDestinationType = checker.getTypeAtLocation(destinationNode);
            // The left hand side of the "as".
            const sourceNode = destinationNode.expression;
            const rawSourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafe(rawDestinationType, rawSourceType)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            }
        };
        return {
            TSTypeAssertion: reportUnsafe,
            TSAsExpression: reportUnsafe,
        };
    },
};
exports.default = noUnsafeTypeAssertion;
